from semantic import *
from earley.feature import *

sem = SemanticRuleSet()
identity = lambda x: x

### Rules ###

# Start rules
sem.add("Start -> S", lambda s: processSentence(s))
sem.add("Start -> Q[wh +]", lambda q: whQuestion(q))
sem.add("Start -> Q[wh -]", lambda q: ynQuestion(q))

# Declarative sentence
sem.add("S -> NP[wh -] VP", lambda np, vp: vp(np))

# Yes-no question
sem.add("Q[wh -] -> Do_Modal NP[wh -] VBAR[fin +]",
	lambda dm, np, vbar: dm(vbar(np)))
# passives
sem.add("Q[wh -] -> Be[tense +] NP[wh -] Vbar.be",
        lambda be, patient, vbar: be(vbar(patient)))

# Where did ...
sem.add("Q[wh +] -> AdvP[wh +] Q[wh -]",
	lambda advp, q:
	q.addFeature(str(advp.getFeature("type").value()),
                     advp.getFeature("var").value()))

sem.add("Q[wh +] -> NP[wh +] VP", 
	lambda np, vp: vp(lambda vbar: vbar(np)))

sem.add("Q[wh +] -> NP[wh +] Q/NP",
	lambda np, q: q(np))

# passives (e.g. "who was an eggplant eaten by")
sem.add("Q/NP -> Be[tense +] NP Vbar.be/NP",
        lambda be, patient, vbar: \
        lambda np: be(vbar(patient)(np)))
sem.add("Vbar.be/NP -> V.part[form V2] PP.by/NP",
        lambda vpart, e: lambda patient: lambda subj: \
        patient(lambda np: vpart(np)(subj)))
sem.add("Vbar.be/NP -> Not Vbar.be/NP",
        lambda neg, vbar: lambda patient: lambda subj: \
        neg(vbar(patient)(subj)))

# non-passive wh-questions
sem.add("Q/NP -> Do_Modal NP VBAR[fin +]/NP",
	lambda dm, np, vbar: lambda obj: dm(np(vbar(obj))))

sem.add("VBAR[fin +]/NP -> V2[tense -] NP/NP",
	lambda v2, e: v2)

sem.add("VBAR[fin +]/NP -> V3[tense -] NP PP.dat/NP",
	lambda v3, np, pp: lambda beneficiary: lambda subj: \
        np(lambda npdo: v3(npdo)(beneficiary)(subj)))

sem.add("VBAR[fin +]/NP -> V3[tense -] NP/NP PP.dat",
	lambda v3, np, pp: lambda patient: lambda subj: \
        pp(v3(patient))(subj))

# (Question 4 - "What did John give Mary")
sem.add("VBAR[fin +]/NP -> V3[tense -] NP NP/NP",
	lambda v3, npio, npdo: lambda patient: lambda subj: \
        npio(v3(patient))(subj))

# (e.g. "who did john not see")
sem.add("VBAR[fin +]/NP -> Not VBAR[fin +]/NP",
        lambda neg, vbar: lambda moved: lambda subj: neg(vbar(moved)(subj)))

# NP
sem.add("NP[pro -, wh -] -> Name", identity)
sem.add("NP[pro -, wh -] -> Det AP* N", lambda det, apstar, n: n(det, apstar))
sem.add("NP[pro -, wh -] -> AP* N[mass +]", lambda apstar, nmass: nmass(apstar))
sem.add("NP[pro -, wh -] -> AP* N[mass -, number plural]",
        lambda apstar, n: n(lambda noun: lambda vp: vp(noun), apstar))

# AdvP[wh +]
sem.add("AdvP[wh +] -> Adv[wh +]", identity)

# AP
sem.add("AP* -> AP* AP", lambda apstar, ap: C(ap, mod=apstar))
sem.add("AP* ->", lambda: StarCategory())
sem.add("AP -> A", identity)

# VP
sem.add("VP -> V+args", lambda v: lambda subj: v(subj))
sem.add("VP -> Do_Modal VBAR[fin +]",
	lambda dm, vbar: lambda subj: dm(vbar(subj)))
# (vbar can be preceded by 'not')
sem.add("VBAR[fin +] -> Not VBAR[fin +]",
        lambda neg, vbar: lambda subj: neg(vbar(subj)))

# passives of transitive verbs
sem.add("V+args -> Be[tense +] Vbar.be",
        lambda be, vbar: lambda patient: be(vbar(patient)))
sem.add("Vbar.be -> V.part[form V2]", lambda vpart: \
        lambda patient: patient(lambda np: vpart(np)(C("Object", name="something"))))
sem.add("Vbar.be -> V.part[form V2] PP.by", lambda vpart, pp: \
        lambda patient: patient(lambda np: pp(lambda x: vpart(np)(x))))
# passives with 'not'
sem.add("Vbar.be -> Not Vbar.be", lambda neg, vbar: \
        lambda patient: neg(vbar(patient)))

# intransitive verbs
sem.add("V+args -> V1[tense +]", lambda v1: lambda subj: subj(v1))
sem.add("V+args -> V1[tense +] PP", lambda v1, pp: lambda subj: pp(subj(v1)))
sem.add("VBAR[fin +] -> V1[tense -]", lambda v1: lambda subj: subj(v1))
sem.add("VBAR[fin +] -> V1[tense -] PP", lambda v1, pp: lambda subj: pp(subj(v1)))

# transitive verbs
sem.add("V+args -> V2[tense +] NP", lambda v2, obj: \
        lambda subj: obj(lambda np: subj(v2(np))))
sem.add("V+args -> V2[tense +] NP[wh -] PP",
	lambda v2, obj, pp: lambda subj: pp(obj(lambda np: subj(v2(np)))))
sem.add("VBAR[fin +] -> V2[tense -] NP[wh -]",
	lambda v2, obj: lambda subj: obj(lambda np: subj(v2(np))))
sem.add("VBAR[fin +] -> V2[tense -] NP[wh -] PP",
	lambda v2, obj, pp: lambda subj: pp(obj(lambda np: subj(v2(np)))))

# ditransitive verbs
sem.add("V+args -> V3[tense +] NP[wh -] PP.dat",
	lambda v3, obj, pp: lambda subj: \
        obj(lambda npdo: pp(lambda npio: subj(v3(npdo)(npio)))))
sem.add("V+args -> V3[tense +] NP[wh -] PP.dat PP",
	lambda v3, obj, pp, ppmod: lambda subj: \
        ppmod(obj(lambda npdo: pp(lambda npio: subj(v3(npdo)(npio))))))
sem.add("VBAR[fin +] -> V3[tense -] NP[wh -] PP.dat",
	lambda v3, obj, pp: lambda subj: \
        obj(lambda npdo: pp(lambda npio: subj(v3(npdo)(npio)))))
sem.add("VBAR[fin +] -> V3[tense -] NP[wh -] PP.dat PP",
	lambda v3, obj, pp, ppmod: lambda subj: \
        ppmod(obj(lambda npdo: pp(lambda npio: subj(v3(npdo)(npio))))))
# (Questions 2 & 3 - "John gave Mary Fido" and "John did give Mary Fido")
sem.add("V+args -> V3[tense +] NP[wh -] NP[wh -]",
	lambda v3, pp, obj: lambda subj: \
        obj(lambda npdo: pp(lambda npio: subj(v3(npdo)(npio)))))
sem.add("V+args -> V3[tense +] NP[wh -] NP[wh -] PP",
	lambda v3, pp, obj, ppmod: lambda subj: \
        ppmod(obj(lambda npdo: pp(lambda npio: subj(v3(npdo)(npio))))))
sem.add("VBAR[fin +] -> V3[tense -] NP[wh -] NP[wh -]",
	lambda v3, pp, obj: lambda subj: \
        obj(lambda npdo: pp(lambda npio: subj(v3(npdo)(npio)))))
sem.add("VBAR[fin +] -> V3[tense -] NP[wh -] NP[wh -] PP",
	lambda v3, pp, obj, ppmod: lambda subj: \
        ppmod(obj(lambda npdo: pp(lambda npio: subj(v3(npdo)(npio))))))

# PP
sem.add("PP[loc ?x] -> P[loc ?x] NP[wh -]", lambda p, np: p(np))
sem.add("PP -> P NP", lambda p, np: p(np))
sem.add("PP.dat -> P.dat NP", lambda p, np: p(np))
sem.add("PP.by -> P.by NP", lambda p, np: p(np))

# Do or Modal
sem.add("Do_Modal -> Do[tense +]", identity)
sem.add("Do_Modal -> Modal", identity)

# Empty slash rules
sem.add("NP/NP ->", lambda: None)
sem.add("PP.dat/NP -> P.dat NP/NP", lambda p, np: None)
sem.add("PP.by/NP -> P.by NP/NP", lambda p, np: None)

### Lexicon ###

# Names
sem.addMatch("Name -> *", lambda name:\
	lambda vp: vp(C("Object", name=name)))

sem.addLexicon("Name",
	['Bob', 'Bill', 'Fido', 'Fred', 'Jane', 'John', 'Mary', 'Poirot', 'Rusty'])

# Common nouns
sem.addMatch("N[mass -] -> *", lambda word:\
	lambda det, apstar:\
		det(C("Object", name=word, mod=apstar)))
sem.addMatch("N[mass +] -> *", lambda word:\
	lambda det, apstar:\
		det(C("Object", name=word, number='mass',
                      mod=apstar)))

sem.addLexicon("N[mass -, number singular]",
	['book', 'cadillac', 'car', 'case', 'cat', 'city', 'communist',
         'conference','corner', 'cow', 'dog', 'drug', 'eggplant', 'exam',
         'fact', 'friend', 'guy', 'horse', 'kennel', 'key', 'letter', 'man',
         'park', 'pencil', 'pot', 'room', 'table', 'tax', 'team', 'theory',
	 'tree', 'tune', 'volume', 'werewolf', 'woman', 'year'])	 
sem.addLexicon("N[mass -, number plural]",
	['books', 'cadillacs', 'cars', 'cases', 'cats', 'cities', 'communists',
         'conferences', 'corners', 'cows', 'dogs', 'drugs', 'eggplants', 'exams',
         'facts', 'friends', 'guys', 'horses', 'kennels', 'keys', 'letters',
         'men', 'mets', 'parks', 'pencils', 'pots', 'redsox', 'rooms', 'teams',
         'theories', 'trees', 'tunes', 'volumes', 'werewolves', 'women', 'years'])
sem.addLexicon("N[mass +]", ['icecream'])

# determiners
sem.add("Det -> 'the'", lambda word: \
        lambda noun: lambda vp: vp(noun))
sem.add("Det -> 'this'", lambda word: \
        lambda noun: lambda vp: vp(noun))
sem.add("Det -> 'a'", lambda word: \
        lambda noun: lambda vp: vp(noun))
sem.add("Det -> 'an'", lambda word: \
        lambda noun: lambda vp: vp(noun))
sem.add("Det -> 'no'", lambda word: \
        lambda noun: lambda vp: opposite(vp(noun)))

# wh-words
sem.addMatch("NP[pro +, wh +] -> *", lambda word: VariableValue(word))
sem.addLexicon("NP[pro +, wh +]", ['who', 'what'])
sem.add("Adv[wh +] -> 'where'", lambda word:\
	C('Adverb', type='locative', var=VariableValue("where")))

def addVerb(form, root, past, present, ppart=None):
	global sem
	if ppart is None: ppart = past
	(pos, proc) = form
	sem.add(CFGProduction(pos.mix(C(None, tense=False)), root),
		proc(root, False))
	sem.add(CFGProduction(pos.mix(C(None, tense=True)), past),
		proc(root, "past"))
	sem.add(CFGProduction(pos.mix(C(None, tense=True)), present),
		proc(root, "present"))
	sem.add(CFGProduction(C("V.part", form=pos), ppart),
		proc(root, "past-participle"))

v1form = (C("V1"),
	lambda root, tense:\
		lambda word: lambda agent:\
			C("Event", action=root, agent=agent, tense=tense,
                          truth="true"))

v2form = (C("V2"),
	lambda root, tense:\
		lambda word: lambda patient: lambda agent: \
			C("Event", action=root, agent=agent, patient=patient,
			tense=tense, truth="true"))

v3form = (C("V3"),
	lambda root, tense:\
		lambda word: lambda patient: lambda beneficiary: lambda agent:\
			C("Event", action=root, agent=agent, patient=patient,
			beneficiary=beneficiary, tense=tense, truth="true"))

addVerb(v1form, 'come', 'came', 'comes', 'come')
addVerb(v1form, 'drive', 'drove', 'drives', 'driven')
addVerb(v1form, 'eat', 'ate', 'eats', 'eaten')
addVerb(v1form, 'return', 'returned', 'returns')
addVerb(v1form, 'sing', 'sang', 'sings', 'sung')
addVerb(v1form, 'sleep', 'slept', 'sleeps')
addVerb(v1form, 'talk', 'talked', 'talks')

addVerb(v2form, 'eat', 'ate', 'eats', 'eaten')
addVerb(v2form, 'buy', 'bought', 'buys')
addVerb(v2form, 'catch', 'caught', 'catches')
addVerb(v2form, 'chase', 'chased', 'chases')
addVerb(v2form, 'compute', 'computed', 'computes')
addVerb(v2form, 'drive', 'drove', 'drives', 'driven')
addVerb(v2form, 'find', 'found', 'finds')
addVerb(v2form, 'hate', 'hated', 'hates')
addVerb(v2form, 'keep', 'kept', 'keeps')
addVerb(v2form, 'kill', 'killed', 'kills')
addVerb(v2form, 'like', 'liked', 'likes')
addVerb(v2form, 'love', 'loved', 'loves')
addVerb(v2form, 'make', 'made', 'makes')
addVerb(v2form, 'pass', 'passed', 'passes')
addVerb(v2form, 'please', 'pleased', 'pleases')
addVerb(v2form, 'sing', 'sang', 'sings', 'sung')
addVerb(v2form, 'see', 'saw', 'sees', 'seen')
addVerb(v2form, 'support', 'supported', 'supports')
addVerb(v2form, 'surprise', 'surprised', 'surprises')
addVerb(v2form, 'visit', 'visited', 'visits')
addVerb(v2form, 'kiss', 'kissed', 'kisses')

addVerb(v3form, 'give', 'gave', 'gives', 'given')

sem.add("P[loc -] -> 'in'",
	lambda word: lambda location: lambda frame:\
	location(lambda np: frame.addFeature("locative",
                C("Place", relation="in", location=np))))

sem.add("P[loc -] -> 'under'",
	lambda word: lambda location: lambda frame:\
	location(lambda np: frame.addFeature("locative",
                C(None, relation="under", location=np))))

sem.add("P[loc +] -> 'on'",
	lambda word: lambda location: lambda frame:\
        location(lambda np: frame.addFeature("locative",
                C(None, relation="on", location=np))))

sem.add("P[loc -] -> 'of'",
	lambda word: lambda source: lambda frame:\
	source(lambda np: frame.addFeature("source", np)))

sem.add("P.dat -> 'to'",
	lambda word: identity)

sem.add("P.by -> 'by'",
	lambda word: identity)

# Adjectives
sem.addMatch("A -> *", lambda word: C(word))
sem.addLexicon("A", ['rabid', 'raw', 'smart', 'red', 'blue', 'green', 'yellow'])

sem.add("A.pred -> 'suspicious'", lambda word: lambda source:\
	C("suspicious", source=source))

# Aux and Modal
sem.add("Do[tense +] -> 'did'",
	lambda word: lambda verb_frame:\
	verb_frame.addFeature("tense", "past"))

sem.add("Do[tense +] -> 'does'",
	lambda word: lambda verb_frame:\
	verb_frame.addFeature("tense", "present"))

sem.add("Be[tense +] -> 'is'",
	lambda word: lambda verb_frame:\
	verb_frame.addFeature("tense", "present"))

sem.add("Be[tense +] -> 'was'",
	lambda word: lambda verb_frame:\
	verb_frame.addFeature("tense", "past"))

sem.add("Modal -> 'can'",
	lambda word: lambda verb_frame:\
	verb_frame.addFeature("mood", "can").addFeature("tense", "present"))

# the meaning of 'not'
sem.add("Not -> 'not'",
        lambda word: lambda verb_frame:\
        opposite(verb_frame))

sem.add("To -> 'to'", lambda word: lambda: None)
sem.add("That -> 'that'", lambda word: lambda: None)
sem.add("Comp[wh +] -> 'whether'", lambda word: lambda: None)
sem.add("NP[wh -, pro +] -> 'everybody'", lambda word: lambda v: v(C("Object"))) ### afk-changed from 'somebody' to 'everybody' (to match the actual meaning)

### Sentence processing ###

def processSentence(data):
        prev = ynQuestion(data)
        if prev == "Yes.":
                return "You already said that."
        if prev == "No.":
                sem.learned.remove(opposite(data))
                sem.learned.add(data)
                return "You said the opposite before, but okay."
        if prev == "I don't know.":
                sem.learned.add(data)
                return "Okay."

def ynQuestion(data):
        # if the database says this statement is true, say Yes
	if len(sem.learned.match(data)) > 0: return "Yes."
	# if the database says this statement is false, say No
        if len(sem.learned.match(opposite(data))) > 0: return "No."
	# otherwise, we don't know
	return "I don't know."

def whQuestion(data):
	matches = sem.learned.matchVar(data)
	if len(matches) == 0: return "I don't know."
	answers = []
	for a in matches:
		if len(a.keys()) == 1:
			answers.append(str(translate(a.values()[0])))
		else:
			answers += [translate(x) for x in sem.learned.match(data)]
	return " and ".join(answers)

def translate(data):
	if data.symbol() == "Object":
		name = data.getFeature("name").value()
		plural = data.getFeature("plural").value()
		mod = translateModifier(data.getFeature("mod").value())
		desc = "%s%s" % (mod, name)
		if name != Category.null():
			return desc
		else: return "something"
	elif data.symbol() == "Place":
		relation = data.getFeature("relation").value()
		location = translate(data.getFeature("location").value())
		return "%s %s" % (relation, location)
	else:
                return repr(data)

def translateModifier(data):
	if data == Category.null(): return ""
	else: return "%s%s " %\
	(translateModifier(data.getFeature("mod").value()), data.symbol())

def oppositeTruth(data):
        """Return "false" if this thing has truth-value "true", otherwise
           return "true"."""
        truthval = repr(data.getFeature("truth").value())
	if truthval == "true":
                return "false"
	else:
                return "true"

def opposite(data):
        """Return the opposite of this thing (i.e. the same thing but with
           the opposite truth-value."""
        return data.addFeature("truth", oppositeTruth(data))
